//
// Copyright 1996 Steve Passe and Paul Crook
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for any purpose and without fee is hereby granted, provided
// that the above copyright notice appear in all copies and that both that
// copyright notice and this permission notice appear in supporting
// documentation, and that the name of AUTHORS (Steve Passe and Paul Crook)
// are not be used in advertising or publicity pertaining to distribution of
// the software without specific, written prior permission.  AUTHORS make no
// representations about the suitability of this software for any purpose.
// It is provided "as is" without express or implied warranty.
//
// AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL AUTHORS
// BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
// OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
// CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//
// 1996 Revision
// AUTHORS:     Steve Passe <smp@csn.net>
//              Paul Crook
//

static char rcsid[] = "$Id: rcclient.cc,v 1.2 1996/03/07 02:00:41 smp Exp smp $";


//
// Remote Control daemon
//

#include <stdio.h>
#include <iostream.h>
#include <termios.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <sys/param.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "util/misc/misc.hh"
#include "util/misc/udgsocket.hh"
#include "util/misc/namedstr.hh"
#include "util/misc/keymap.hh"
#include "util/misc/tokens.hh"
#include "rcclient.hh"
#include "rcserver.hh"


//
// Note:
//	we are assumming that there is ONE AND ONLY ONE instance of
//	an RcClient object per process.  This makes signal handling,
//	etc. easier for now.  In the future we will need to think thru
//	a multithreaded model...
//

//
// base constructor
//
RcClient::RcClient( const char* name )
	:
	 _myName( name )
{
    struct sigaction	action;

    // prepare for signals
    sigemptyset( &action.sa_mask );
    action.sa_flags = 0;

    // setup the INTerrupt handler
    /// FIXME: real issue is version of g++ used(???)
#if 0 /**(defined(BSD) && (BSD >= 199306))*/
    action.sa_handler = signalTerminate;
#else
    action.sa_handler = (PFIV)signalTerminate;
#endif  // BSD

    //
    _verbose = 0;

#if defined( HAVE_VOICE )
    _verbal = FALSE;
#endif  // HAVE_VOICE

    // load defaults file
    /// not sure this should be here as opposed to initialize(),
    /// but it is helpful to be here for several reasons
    if ( _defaults.readStrings( RCD_STRINGS ) == ERROR )
    {
	cerr << form( "couldn't load STRINGS file: %s", RCD_STRINGS ) << endl;
    }

    return;
}


//
// 
//
int
RcClient::initialize( const char* pathName,	// symbolic path name
		      const char* logName )	// symbolic log name
{
    // (optionally) create a logfile for stderr
    if ( logName )
    {
	// get logfile name
	const char* p;
	if ( (p = _defaults.getString( logName )) == NULL )
	{
	    cerr << form( "couldn't open log file: %s", logName ) << endl;
	    return ERROR;
	}

	// reopen 'stderr' as log file
	if ( freopen( p, "a+", stderr ) == NULL )
	{
	    perror( "logfile open()" );
	    return ERROR;
	}
	setbuf( stderr, NULL );    // no buffer of output
    }

    // setup data for select()
    _clientFD = fileno( stdin );
    FD_ZERO( &_selectSet );
    FD_SET( _clientFD, &_selectSet );
    _maxFD = _clientFD;

    if ( pathName )
    {
	// get the socket paths from the RC defaults file
	const char*	itsPath;
	if ( !(_myPath = _defaults.getString( "RCD_DG_TMP" ))
	    || !(itsPath = _defaults.getString( pathName )) )
	{
	    cerr << "couldn't get socket paths\n";
	    return ERROR;
	}

	// create a socket
	_daemonSocketPtr = new UnixDGSocket( (char*)_myPath, (char*)itsPath );
	if ( (_daemonFD = _daemonSocketPtr->getFD()) == ERROR )
	{
	    return ERROR;
	}
	FD_SET( _daemonFD, &_selectSet );
	_maxFD = MAX( _maxFD, _daemonFD );
    }

#if defined( USES_KEYMAP )
    // find and read the command database
    const char*	s = _defaults.getString( "BIN???KEYMAP" );
    if ( s == NULL )
    {
	cerr << "couldn't find KEYMAP" << endl;
	return ERROR;
    }

    /// if ...
    if ( _Map.readMap( (char*)s ) == ERROR )
    {
	cerr << "couldn't find ??? " << s << endl;
	return;
    }
#endif  // USES_KEYMAP

    // everything worked
    return OK;
}


//
// wait for IO from either the device port/pipe or the client socket(s):
//
int
RcClient::waitForData( void )
{
    int		count, status = 0;

    // wait for traffic
    _fdSet = _selectSet;

    // wait for commands/data from clients or devices
    count =
      select( _maxFD+1, &_fdSet, (fd_set*)0, (fd_set*)0, (struct timeval*)0 );

    // ERROR
    if ( count < 0 )
    {
	return ERROR;
    }

    // device wants attention
    if ( FD_ISSET( _daemonFD, &_fdSet ) )
    {
	status = DAEMON_DATA;
    }

    // client wants attention
    if ( FD_ISSET( _clientFD, &_fdSet ) )
    {
	status |= CLIENT_DATA;
    }

    return status;
}


//
// wait for IO from multiple sources
//
int
RcClient::waitOnData( void )
{
    // setup select() input fd_set
    _fdSet = _selectSet;

    // wait for commands/data from sources
    int count =
      select( _maxFD+1, &_fdSet, (fd_set*)0, (fd_set*)0, (struct timeval*)0 );

    // return  result
    return (count < 0) ? ERROR : count;
}


//
// 
//
void
RcClient::addSelect( int fd )
{
    FD_SET( fd, &_selectSet );
    _maxFD = MAX( _maxFD, fd );
}


//
// 
//
void
RcClient::rmvSelect( int fd )
{
    int max = 0;

    // clear the fd from set
    FD_CLR( fd, &_selectSet );

    // find the new 'max'
    if ( fd == _maxFD )
    {
	for ( int x = 0; x <= _maxFD; ++x )
	{
	    if ( FD_ISSET( x, &_selectSet ) ) max = x;
	}
	_maxFD = max;
    }
}


//
// make a User DataGram Connection
//
int
RcClient::makeUDGConnection( const char* destination,
			     Connection*& connection,
			     int& connectionFD,
			     char* tmpAddr )
{
    // convert RC pathstring to 'well known address'
    const char*	destAddr;
    if ( (destAddr = _defaults.getString( destination )) == 0 )
    {
	cerr << "makeUDGConnection(): couldn't get path" << endl;
	return ERROR;
    }

    // create and (optionally) return a process# based address
    char myAddr[] = UDG_TMPADDR;
    mktemp( myAddr );
    if ( tmpAddr )
    {
	strncpy( tmpAddr, myAddr, UDG_TMPADDR_SIZE );
    }

    // build a connection
    connection = new Connection( (const char*)myAddr, destAddr );

    // test the connection
    connectionFD = connection->getFD();
    if ( connectionFD == ERROR )
    {
	cerr << "makeUDGConnection(): couldn't create Connection" << endl;
	return ERROR;
    }

    // add fd to select() fd_set
    addSelect( connectionFD );

    // done
    return OK;
}


//
// read from a client, saving return address
//
int
RcClient::readDaemonMsg( char* msg )
{
    int count;

    // read from socket, saving the senders address in _retAddr
    if ( (count = _daemonSocketPtr->recv( msg )) < 0 )
    {
	perror( "readDaemonMsg(): recv()" );
	return ERROR;
    }

    // effective EOF, time to quit
    if ( count == 0 ) cerr << "ignoring EOF" << endl; //return FALSE;

    return count;
}


//
// write to a client @ most recent 'readDaemonMsg()' address
//
int
RcClient::writeDaemonMsg( const char* msg, int count )
{
    // return count bytes in msg
    if ( _daemonSocketPtr->send( msg, count ) != count )
    {
	perror( "writeDaemonMsg(): send()" );
	return ERROR;
    }

    return count;
}


#if 0

/// FIXME: make more efficient
//
// split message into path & msg, create a connection object and send message
//
int
RcClient::sendMessage( const char* msg )
{
    const char*	myPath;
    const char*	itsPath;

    // make a working copy
    char* args = new char[ strlen( msg ) + 1 ];
    strcpy( args, msg );

    // break string up into address & message
    Tokens& tokens = *new Tokens( (char*)args, 2 );
    int argc = tokens.count();

    // create an Message connection
    myPath = _defaults.getString( "RCD_DG_TMP" );
    itsPath = _defaults.getString( tokens[ 0 ] );

    if ( (!myPath) || (!itsPath) )
    {
	delete &tokens;
	cerr << "couldn't get socket paths\n";
	return ERROR;
    }

    // create a connection and send message to the daemon
    Connection& connection = *new Connection( myPath, itsPath );
    connection.send( tokens[ 1 ] );

    // return memory
    delete &connection;
    delete &tokens;
    delete args;

    return OK;
}


/// FIXME: make more efficient
//
// create a connection object and send message
//
int
RcClient::sendMessage( const char* path, const char* msg )
{
    const char*	myPath;
    const char*	itsPath;

    // create an Message connection
    myPath = _defaults.getString( "RCD_DG_TMP" );
    itsPath = _defaults.getString( path );

    if ( (!myPath) || (!itsPath) )
    {
	cerr << "couldn't get socket paths\n";
	return ERROR;
    }

    // create a connection and send message to the daemon
    Connection& connection = *new Connection( myPath, itsPath );
    connection.send( msg );

    // return memory
    delete &connection;

    return OK;
}


/// FIXME: make more efficient
//
// create a connection object, send message, receive reply
//
int
RcClient::sendMessage( const char* path, const char* msg, char* reply )
{
    const char*	myPath;
    const char*	itsPath;

    // create an Message connection
    myPath = _defaults.getString( "RCD_DG_TMP" );
    itsPath = _defaults.getString( path );

    if ( (!myPath) || (!itsPath) )
    {
	cerr << "couldn't get socket paths\n";
	return ERROR;
    }

    // create a connection and send message to the daemon
    Connection& connection = *new Connection( myPath, itsPath );
    connection.send( msg );
    connection.recv( reply );

    // return memory
    delete &connection;

    return OK;
}
#endif  // 0


//
// 
//
void
RcClient::terminate( void )
{
    char	s[ 256 ];

    sprintf( s, "\n\n%s daemon stopped: ", _myName );
    markLog( s );

    exit( 0 );
}


//
// termination caused by signal
//  this function is static to allow SIGNALs to call it
//
void
RcClient::signalTerminate( void )
{
/// FIXME:
/// in destructor???
///    unlink( RCD_DG_PATH );

    cerr << "\n\nclient stopped via SIGNAL\n" << endl;

    exit( 1 );
}


//
// place a time-stamped message into log file
//
void
RcClient::markLog( const char* m )
{
    char	s[ 255 ];
    time_t	clock;

    time( &clock );
    strcpy( s, m );
    strcat( s, ctime( &clock ) );
    cerr << s << endl;
}
