//
// Copyright 1996 Steve Passe and Paul Crook
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for any purpose and without fee is hereby granted, provided
// that the above copyright notice appear in all copies and that both that
// copyright notice and this permission notice appear in supporting
// documentation, and that the name of AUTHORS (Steve Passe and Paul Crook)
// are not be used in advertising or publicity pertaining to distribution of
// the software without specific, written prior permission.  AUTHORS make no
// representations about the suitability of this software for any purpose.
// It is provided "as is" without express or implied warranty.
//
// AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL AUTHORS
// BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
// OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
// CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//
// 1996 Revision
// AUTHORS:     Steve Passe <smp@csn.net>
//              Paul Crook
//

static char rcsid[] = "$Id: rcdaemon.cc,v 1.6 1996/03/05 03:54:41 smp Exp smp $";


//
// Remote Control daemon
//

#include <stdio.h>
#include <iostream.h>
#include <termios.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <sys/param.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "util/misc/misc.hh"
#include "util/misc/udgsocket.hh"
#include "util/misc/namedstr.hh"
#include "util/misc/keymap.hh"
#include "util/misc/tokens.hh"
#include "rcdaemon.hh"
#include "rcserver.hh"


//
// Note:
//	we are assumming that there is ONE AND ONLY ONE instance of
//	an RcDaemon object per process.  This makes signal handling,
//	etc. easier for now.  In the future we will need to think thru
//	a multithreaded model...
//

//
// base constructor
//
RcDaemon::RcDaemon( void )
{
    struct sigaction	action;

    // prepare for signals
    sigemptyset( &action.sa_mask );
    action.sa_flags = 0;

    // setup the INTerrupt handler
    /// FIXME: real issue is version of g++ used
#if 0 /**(defined(BSD) && (BSD >= 199306))*/
    action.sa_handler = signalTerminate;
#else
    action.sa_handler = (PFIV)signalTerminate;
#endif  // BSD

    sigaction( SIGINT, &action, (struct sigaction *)NULL );
    sigaction( SIGQUIT, &action, (struct sigaction *)NULL );
    sigaction( SIGABRT, &action, (struct sigaction *)NULL );

#if defined( NO_ZOMBIES_ALLOWED )
    action.sa_handler = childDeath;
    sigaction( SIGCHLD, &action, (struct sigaction *)NULL );
#endif  // NO_ZOMBIES_ALLOWED

    // close unneeded files
    fclose( stdin );
    fclose( stdout );

    //
    _verbose = 0;

#if defined( HAVE_VOICE )
    _verbal = FALSE;
#endif  // HAVE_VOICE

    // load defaults file
    /// not sure this should be here as opposed to initialize(),
    /// but it is helpful to be here for several reasons
    if ( _defaults.readStrings( RCD_STRINGS ) == ERROR )
    {
	cerr << form( "couldn't load STRINGS file: %s", RCD_STRINGS ) << endl;
    }

    return;
}


//
// 
//
int
RcDaemon::initialize( const char* name,		// process name
		      const char* pathName,	// symbolic path name
		      const char* logName,	// symbolic log name
		      const int port )		// device fd
{
    // 
    _myName = name;

    // get logfile name
    const char*		p;
    if ( (p = _defaults.getString( logName )) == NULL )
    {
	cerr << form( "couldn't open log file: %s", logName ) << endl;
	return ERROR;
    }

    // reopen 'stderr' as log file
    if ( freopen( p, "a+", stderr ) == NULL )
    {
	perror( "logfile open()" );
	return ERROR;
    }
    setbuf( stderr, NULL );    // no buffer of output

    // setup socket for client messages
    if ( (_myPath = _defaults.getString( pathName )) == NULL )
    {
	cerr << form( "couldn't read my socketpath: %s", pathName ) << endl;
	return ERROR;
    }
    _clientSocketPtr = new UnixDGSocket( (char*)_myPath );

#if defined( USES_KEYMAP )
    // find and read the command database
    const char*	s = _defaults.getString( "BIN???KEYMAP" );
    if ( s == NULL )
    {
	cerr << "couldn't find KEYMAP" << endl;
	return ERROR;
    }

    /// if ...
    if ( _Map.readMap( (char*)s ) == ERROR )
    {
	cerr << "couldn't find ??? " << s << endl;
	return;
    }
#endif  // USES_KEYMAP

    // announce ourselves
    char s[ 255 ];
    sprintf( s, "\n\n%s daemon started: ", _myName );
    logTime( s );

    // setup data for select()
    FD_ZERO( &_fdSet );
    if ( (_clientFD = _clientSocketPtr->getFD()) < 0 )
    {
	return ERROR;
    }
    _deviceFD = port;
    _maxFD = MAX( port, _clientFD ) + 1;

    // everything worked
    return OK;
}


//
// wait for IO from either the device port/pipe or the client socket(s):
//
int
RcDaemon::waitForData( void )
{
    int		count, status = 0;

    // wait for traffic
    FD_SET( _clientFD, &_fdSet );
    FD_SET( _deviceFD, &_fdSet );

    // wait for commands/data from clients or devices
    count =
      select( _maxFD, &_fdSet, (fd_set*)0, (fd_set*)0, (struct timeval*)0 );

    // ERROR
    if ( count < 0 )
    {
	return ERROR;
    }

    // device wants attention
    if ( FD_ISSET( _deviceFD, &_fdSet ) )
    {
	status = DEVICE_DATA;
    }

    // client wants attention
    if ( FD_ISSET( _clientFD, &_fdSet ) )
    {
	if ( readClientMsg() == ERROR )
	{
	    terminate();
	}
	status |= CLIENT_DATA;
    }

    return status;
}


//
// read from a client, saving return address
//
int
RcDaemon::readClientMsg( void )
{
    int count;

    // read from socket, saving the senders address in _retAddr
    if ( (count = _clientSocketPtr->recv( _clientMsg, &_retAddr )) < 0 )
    {
	perror( "readClientMsg(): recv()" );
	return ERROR;
    }

    // effective EOF, time to quit
    if ( count == 0 ) cerr << "ignoring EOF" << endl; //return FALSE;

    return OK;
}


//
// write to a client @ most recent 'readClientMsg()' address
//
int
RcDaemon::writeClientMsg( const char* msg, int count )
{
    // return count bytes in msg
    if ( _clientSocketPtr->send( msg, count, &_retAddr ) != count )
    {
	perror( "writeClientMsg(): send()" );
	return ERROR;
    }

    return OK;
}


//
// write to a client @ address
//
int
RcDaemon::writeClientMsg( const char* msg, int count, UNIXSOCKADDR& addr )
{
///cerr << form( "p: '%s', l: %d", addr.addr.sun_path, addr.len ) << endl;
///cerr << form( "p: '%s', l: %d", _retAddr.addr.sun_path, _retAddr.len ) << endl;
    // return count bytes in msg
    if ( _clientSocketPtr->send( msg, count, &addr ) != count )
    {
	perror( "writeClientMsg(): send()" );
	return ERROR;
    }

    return OK;
}


/// FIXME: make more efficient
//
// split message into path & msg, create a connection object and send message
//
int
RcDaemon::sendMessage( const char* msg )
{
    const char*	myPath;
    const char*	itsPath;

    // make a working copy
    char* args = new char[ strlen( msg ) + 1 ];
    strcpy( args, msg );

    // break string up into address & message
    Tokens& tokens = *new Tokens( (char*)args, 2 );
    int argc = tokens.count();

    // create an Message connection
    myPath = _defaults.getString( "RCD_DG_TMP" );
    itsPath = _defaults.getString( tokens[ 0 ] );

    if ( (!myPath) || (!itsPath) )
    {
	delete &tokens;
	cerr << "couldn't get socket paths\n";
	return ERROR;
    }

    // create a connection and send message to the daemon
    Connection& connection = *new Connection( myPath, itsPath );
    connection.send( tokens[ 1 ] );

    // return memory
    delete &connection;
    delete &tokens;
    delete args;

    return OK;
}


/// FIXME: make more efficient
//
// create a connection object and send message
//
int
RcDaemon::sendMessage( const char* path, const char* msg )
{
    const char*	myPath;
    const char*	itsPath;

    // create an Message connection
    myPath = _defaults.getString( "RCD_DG_TMP" );
    itsPath = _defaults.getString( path );

    if ( (!myPath) || (!itsPath) )
    {
	cerr << "couldn't get socket paths\n";
	return ERROR;
    }

    // create a connection and send message to the daemon
    Connection& connection = *new Connection( myPath, itsPath );
    connection.send( msg );

    // return memory
    delete &connection;

    return OK;
}


/// FIXME: make more efficient
//
// create a connection object, send message, receive reply
//
int
RcDaemon::sendMessage( const char* path, const char* msg, char* reply )
{
    const char*	myPath;
    const char*	itsPath;

    // create an Message connection
    myPath = _defaults.getString( "RCD_DG_TMP" );
    itsPath = _defaults.getString( path );

    if ( (!myPath) || (!itsPath) )
    {
	cerr << "couldn't get socket paths\n";
	return ERROR;
    }

    // create a connection and send message to the daemon
    Connection& connection = *new Connection( myPath, itsPath );
    connection.send( msg );
    connection.recv( reply );

    // return memory
    delete &connection;

    return OK;
}


//
// 
//
void
RcDaemon::terminate( void )
{
    char	s[ 256 ];

    sprintf( s, "\n\n%s daemon stopped: ", _myName );
    logTime( s );

    exit( 0 );
}


//
// termination caused by signal
//  this function is static to allow SIGNALs to call it
//
void
RcDaemon::signalTerminate( void )
{
/// FIXME:
/// in destructor???
///    unlink( RCD_DG_PATH );

    cerr << "\n\ndaemon stopped via SIGNAL\n" << endl;

    exit( 1 );
}


//
// place a time-stamped message into log file
//
void
RcDaemon::logTime( const char* m )
{
    char	s[ 255 ];
    time_t	clock;

    time( &clock );
    strcpy( s, m );
    strcat( s, ctime( &clock ) );
    cerr << s << endl;
}


///////////////////////////////////////////////////////////////////////////////
#if defined( USES_KEYMAPS )
/// FIXME: useful???

    virtual int		send( const char* ) { return -1; }

    virtual int		send( const char* );
    virtual int		command( char* );

//
// break string up into individual commands,
// strip comments,
// and send each via command()
//
int
Hcs2Daemon::send( const char* commands )
{
    // get a local copy of commands, Tokens corrupts it!
    char* s = new char[ strlen( commands ) + 1 ];
    strcpy( s, commands );

    // break string up into individual words
    Tokens& tokens = *new Tokens( s );
    int argc = tokens.count();

    for ( int x = 0; x < argc;  )
    {
	// filter comments
	if ( tokens[x][0] == '#' )
	{
	    x = argc;
	    break;
	}

	command( tokens[ x ] );
	if ( ++x < argc ) sleep( 1 );
    }

    // return storage to system
    delete &tokens;
    delete s;

    return OK;
}


//
// log command and send its associated data to hardware
//
int
Hcs2Daemon::command( char* command )
{
#if defined( USES_KEYMAP )
    // get the ???
    short sample[ 1024 ];
    _hcs2Map.getKey( command, sample );
    int count = (int)sample[ 0 ];
#endif  // USES_KEYMAP

    cerr << form( "-> %-20s", command );
    time_t	clock;
    time( &clock );
    cerr << ctime( &clock );

    ///write( _port, &, 1 );

    return OK;
}
#endif  // USES_KEYMAPS
