//
// Copyright 1996 Steve Passe
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for any purpose and without fee is hereby granted, provided
// that the above copyright notice appear in all copies and that both that
// copyright notice and this permission notice appear in supporting
// documentation, and that the name of AUTHORS (Steve Passe and Paul Crook)
// are not be used in advertising or publicity pertaining to distribution of
// the software without specific, written prior permission.  AUTHORS make no
// representations about the suitability of this software for any purpose.
// It is provided "as is" without express or implied warranty.
//
// AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL AUTHORS
// BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
// OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
// CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//
// 1996 Revision
// AUTHORS:     Steve Passe <smp@csn.net>
//

// $Id: hcs2daemon.hh,v 1.13 1996/03/23 18:35:48 smp Exp smp $


//
// hcs2daemon.hh, header for hcs2daemon.cc
//


#ifndef _HCS2DAEMON_H
#define _HCS2DAEMON_H

#include <unistd.h>
#include "SLList.h"

#include "../daemon/rcdaemon.hh"


#define HCS2_COMMAND		'!'
#define HCS2_REPLY		'$'
#define HCS2_ACK		'^'
#define HCS2_BAD_FIRMWARE	'#'
#define HCS2_TOO_BIG		'@'

#define HCS2_CMD_RESET		0x00
#define HCS2_CMD_PAUSE_STATUS	0x01
#define HCS2_CMD_REQUEST_STATUS	0x02
#define HCS2_CMD_TIME_DISPLAY	0x03

#define HCS2_CMD_LOAD_PROG	0x06

#define HCS2_CMD_DI_DISPLAY	0x13
#define HCS2_CMD_GET_DI_ONCE	0x14
#define HCS2_CMD_DO_DISPLAY	0x16
#define HCS2_CMD_GET_DO_ONCE	0x17
#define HCS2_CMD_ADC_DISPLAY	0x19

#define HCS2_FIRST_ASYNC	0x80
#define HCS2_LAST_ASYNC		0x88
#define HCS2_STAT_DI		0x82
#define HCS2_STAT_DO		0x83
#define HCS2_STAT_AI		0x84
#define HCS2_STAT_AO		0x85
#define HCS2_STAT_CONSOLE	0x88

#define MAXINPUTBUFFER		1024
#define MAX_DATA		256

// returned ASCII messages
#define FAIL_MSG		"<FAIL>"
#define OK_MSG			"<OK>"

//
// IO 'watching' macros
//
#define HCS2_IO_BITS		256
#define HCS2_IO_BLOCKS		(HCS2_IO_BITS/8)

#define TYPE_INPUT		0
#define TYPE_OUTPUT		1

// convert bit# to blockmap byte# (0 thru 30
#define BITMAP_BYTE(X)		(X / 64)

// convert bit# to mask for its blockmap
#define BITMAP_BIT(X)		(1 << ((X % 64) / 8))

// map bit# (0 thru 255) to block# (0 thru 32)
#define IO_BLOCK(X)		(X / 8)

// get bit#_within_block from bit#
#define IO_BLOCKBIT(X)		(X % 8)

// convert bit# to mask for its block
#define IO_BLOCKMASK(X)		(1 << (X % 8))

// convert group# & block_bit# into bit#
#define GROUPBIT_TO_BIT(X,Y)	((X * 8) + Y)


class Hcs2Daemon : public RcDaemon
{

 public:

    // constructor
    Hcs2Daemon( void ) { _port = -1; }

    // destructor
    virtual		~Hcs2Daemon( void ) { cerr <<
					      "\nhcs2Daemon stopped" << endl; }

    // initialize daemon
    virtual int		initialize( const char* name,
				    const char* path,
				    const char* logFile,
				    const int port );

    // initialize device controlled by daemon
    virtual int		deviceInitialize( void );

    //
    // unique HCSII commands & data
    //

    // main processing loop
    int			hcs2Main( void );

    // command to download a program to the HCSII
    int			hcs2LoadProgram( const char* name );

 protected:

    // HCSII specific method for getting data
    virtual int		getDeviceData( void ) { return hcs2Packet( -1 ); }

    // HCSII specific method for termination
    virtual void	terminate( void );

    // HCSII specific method for opening control channel
    virtual int		openDevice( void );


 private:

    //
    // unique HCSII commands & data
    //

    // process client messages
    int			doClient( void );

    // send HCSII commands, various front ends
    int			hcs2Command( char* asciiBytes );
    int			hcs2Command( int cmd, int count ... );

    // send HCSII commands, actual routine
    int			hcs2Command( int count, unsigned char* data );

    // get HCSII packets
    int			hcs2Packet( int command );

    // handle async data generated by HCSII
    int			hcs2AsyncData( int command );

    // read next byte from HCSII
    int			hcs2Byte( void ) { return read( _port, &_byte, 1 ); }
    unsigned char	_byte;

    // process HCSII data
    int			logData( int count, const unsigned char* data );
    int			formatData( int count, const unsigned char* data,
				     char* buffer );
    int			processData( int count, unsigned char* data );
    int			processString( const char* string );
    unsigned char	_data[ MAX_DATA ];

    // HCSII device port
    int			_port;

    // turn off all EXPRESS default reports
    void		hcs2Quiet( void );


    //
    // routines/data for 'watching' input/outputs
    //

    // get the current status of 'bit', update cache byte
    int			updateCurrentIO( int type, int bit );


    // changed 'watched' state of 'bit'
    int			modifyWatchedIO( int type,
					 int what,
					 const char* bit,
					 UNIXSOCKADDR& addr );

    // delete an address from list
    int			deleteFromList( SLList<UNIXSOCKADDR>& list,
				        UNIXSOCKADDR& addr );

    // print a list of addresses
    void		printList( SLList<UNIXSOCKADDR>& list );

    // report changed status
    int			reportChangedIO( int type, int group, u_char state );

    // 1 bit for each input
    unsigned char	_currentInputs[ HCS2_IO_BLOCKS ];

    // 1 bit for each output
    unsigned char	_currentOutputs[ HCS2_IO_BLOCKS ];

    // 1 bit for each input block
    unsigned char	_currentInputMap[ HCS2_IO_BLOCKS / 8 ];

    // 1 bit for each output block
    unsigned char	_currentOutputMap[ HCS2_IO_BLOCKS / 8 ];

    // 1 bit for each input currently being 'watched'
    unsigned char	_watchedInputs[ HCS2_IO_BLOCKS ];

    // 1 bit for each output currently being 'watched'
    unsigned char	_watchedOutputs[ HCS2_IO_BLOCKS ];

    // linked list of clients with pending requests
    SLList<UNIXSOCKADDR> _clientsWatchingInputs[ HCS2_IO_BITS ];
    SLList<UNIXSOCKADDR> _clientsWatchingOutputs[ HCS2_IO_BITS ];


    // state table of HCSII command/data lengths, other info
    class Hcs2CmdInfo
    {

    public:

	Hcs2CmdInfo( void ) {}
	~Hcs2CmdInfo( void ) {}

	int	cmdLength( int command ) { return _entry[ command ].cl; }
	int	dataLength( int command ) { return _entry[ command ].dl; }


    private:

	//
	// meanings:
	//  cl: # of command bytes
	//  dl: # of reply bytes
	//   0 -> 8 == expected length
	//  -1      == variable length
	//  -2      == <command not defined>
	//
	static struct ENTRY {
	    const int	cl;
	    const int	dl;
	} _entry[ 50 ];

    };

    // instance of HCSII command info 
    Hcs2CmdInfo		_cmdInfo;
};

#if 0  // moved to rcdaemon base class
    inline int		returnMessage( const char* msg );
    inline int		returnMessage( const char* msg, UNIXSOCKADDR& addr );
    int			returnFailMsg( void ) {return returnMessage(FAIL_MSG);}
    int			returnOkMsg( void ) { return returnMessage( OK_MSG ); }
#endif  // moved

#endif // _HCS2DAEMON_H
