//
// Copyright 1996 Steve Passe
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for any purpose and without fee is hereby granted, provided
// that the above copyright notice appear in all copies and that both that
// copyright notice and this permission notice appear in supporting
// documentation, and that the name of AUTHORS (Steve Passe and Paul Crook)
// are not be used in advertising or publicity pertaining to distribution of
// the software without specific, written prior permission.  AUTHORS make no
// representations about the suitability of this software for any purpose.
// It is provided "as is" without express or implied warranty.
//
// AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL AUTHORS
// BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
// OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
// CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//
// 1996 Revision
// AUTHORS:     Steve Passe <smp@csn.net>
//

static char rcsid[] = "$Id: mlbusmaster.cc,v 1.5 1996/03/31 03:09:40 smp Exp smp $";


//
// mlbusmaster.cc
//

#include <iostream.h>
#include <string.h>
#include <stdio.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <machine/cpufunc.h>

#include "mlbusmaster.hh"
#include "mldevice.hh"


//
// constructor
//
MLBusMaster::MLBusMaster( int port )
{
    /// NOTE: this will only support ONE busmaster hardware device!
    // initialize delay function
    struct sigaction action;
    sigemptyset( &action.sa_mask );
    action.sa_flags = SA_RESTART;

    // cock timeout timer
    action.sa_handler = (void(*)(int))delayDone;
    if ( sigaction( SIGALRM, &action, NULL ) < 0 )
    {
	cerr << "bad delay()" << endl;
	exit( 1 );
    }

    // determine hardware address of port
    switch ( port )
    {
    case 0:	_ttyAddr = TTY0; break;
    case 1:	_ttyAddr = TTY1; break;
    case 2:	_ttyAddr = TTY2; break;
    case 3:	_ttyAddr = TTY3; break;
    default:
	cerr << "bad port number: " << port << endl;
	exit( 1 );
    }

    // 
    // initialize the port
    IER( 0 );				// disable interrupts
    LCR( 0 );				// necessary???
    MCR( 1 );				// RTS low, DTR high

    // 
    mlReset();
}


//
// reset the MicroLAN(tm) hardware
//
int
MLBusMaster::mlReset( void )
{
    // setup uart
    LCR( BAUD_LOAD );			// enable BAUD load
    BAUD( BAUD_9600 );			// RESET pulse timing
    LCR( DATA_LOAD );			// enable data load
    DATA( RESET_PULSE );

    delay( RESET_TIME );

    // set timing for data
    LCR( BAUD_LOAD );			// enable BAUD load
    BAUD( BAUD_115200 );		// data timing
    LCR( DATA_LOAD );			// enable data load

    // clear "echoed" data
    while ( LSR() & DATA_READY )
    {
	DATA();
    }

    /// FIXME: is lack of echoed data a 'lack of presence'???
    return OK;
}


//
// read contents of a device ROM memory (family, serial#, crc)
//
int
MLBusMaster::readROM( char* asciiBuffer )
{
    const BYTE	command = MLDevice::ROM_READ();
    BYTE	data[ 8 ];

    if ( mlCommand( 1, &command, 8, data ) == ERROR )
    {
	return ERROR;
    }

    sprintf( asciiBuffer, "%02x%02x%02x%02x%02x%02x%02x%02x",
	     data[ 7 ], data[ 6 ], data[ 5 ], data[ 4 ],
	     data[ 3 ], data[ 2 ], data[ 1 ], data[ 0 ] );

    return OK;
}


//
// address device by family, serial#, crc
//
int
MLBusMaster::matchROM( const BYTE* id )
{
    // send Match ROM command
    mlROMCommand( MLDevice::ROM_MATCH() );

    // send id
    if ( mlWriteData( MLDevice::ROM_LENGTH(), id ) == ERROR ) return ERROR;

    return OK;
}


//
// skip the ROM addressing cycle
//
/// FIXME: inline
int
MLBusMaster::skipROM( void )
{
    return ( mlROMCommand( MLDevice::ROM_SKIP() ) == ERROR ) ? ERROR : OK;
}


//
// search for all existing devices
//
//    BYTE* ids = new BYTE[ MLDevice::ROM_LENGTH() * maxDevices ];
//    int number = searchROM( ids, maxDevices );
//

#define setRomBit(I)	(idp[ ((I-1)/8) ] |= (BYTE)(1 << ((I-1)%8)))
#define getRomBit(I)	((idp[ ((I-1)/8) ] & (BYTE)(1 << ((I-1)%8))) ? 1 : 0)
#define getPrevRomBit(I) (((idp-8)[((I-1)/8)] & (BYTE)(1 << ((I-1)%8))) ? 1:0)

int
MLBusMaster::searchROM( BYTE* ids, const int maxID )
{
    BYTE*	idp = ids;
    int		lastDiscrepancy = 0;
    int		discrepancyMarker;
    int 	bitA, bitB;
    int		number;

    for ( number = 0; number < maxID; ++number, idp += 8 )
    {
	// clear ROM area
	memset( idp, 0, 8 );

	// reset the bus for new pass
	if ( mlReset() == ERROR ) return ERROR;

	// send the ROM command to start search
	if ( mlROMCommand( MLDevice::ROM_SEARCH() ) == ERROR ) return ERROR;

	// check each of the 64 ROM bit positions
	discrepancyMarker = 0;
	for ( int romBitIndex = 1; romBitIndex <= 64; ++romBitIndex )
	{
	    // read 'true' and 'comp' bits
	    if ( (bitA = mlReadBit()) == ERROR) return ERROR;
	    if ( (bitB = mlReadBit()) == ERROR) return ERROR;

	    // check for 'no connected devices'
	    if ( (bitA == 1) && (bitB == 1) ) return ERROR;

	    // check for multiple connected devices in current bit position
	    if ( (bitA == 0) && (bitB == 0) )
	    {
		if ( romBitIndex == lastDiscrepancy )
		{
		    setRomBit( romBitIndex );
		}
		else if ( romBitIndex > lastDiscrepancy )
		{
		    discrepancyMarker = romBitIndex;
		}

		// romBitIndex < lastDiscrepancy, ie we've been this way before
		else
		{
		    // get romBit from previous pass
		    int b = getPrevRomBit( romBitIndex );

		    // was zero, do 1st pass of 2-pass cycle
		    if ( b == 0 )
		    {
			discrepancyMarker = romBitIndex;
		    }

		    // was one, do 2nd pass of 2-pass cycle
		    else
		    {
			setRomBit( romBitIndex );
		    }
		}
	    }
	    else
	    {
		if ( bitA == 1 ) setRomBit( romBitIndex );
	    }

	    // send 
	    mlWriteBit( getRomBit( romBitIndex ) );
	}

#if 1
	char asciiBuffer[ 20 ];
	sprintf( asciiBuffer, "%02x%02x%02x%02x%02x%02x%02x%02x",
		idp[7], idp[6], idp[5], idp[4],
		idp[3], idp[2], idp[1], idp[0] );
	cerr << "found device: " << asciiBuffer << endl;
#endif

	lastDiscrepancy = discrepancyMarker;
	if ( lastDiscrepancy == 0 ) break;
    }

    // account for last device
    return (number + 1);
}


//
// ROM function command:
//  bus RESET
//  ROM command
//
int
MLBusMaster::mlROMCommand( const int command )
{
    int	x;

    // 
    for ( x = 0; x < COMMAND_RETRIES; ++x )
    {
	mlReset();

	if ( mlWriteByte( (const BYTE)command ) != ERROR ) break;
    }

    // done
    return (x < COMMAND_RETRIES) ? OK : ERROR;
}


//
// 
//
int
MLBusMaster::mlReadFunction( int function, int dataLength, BYTE* data )
{
    // send command
    if ( mlWriteByte( (const BYTE)function ) == ERROR )
    {
	return ERROR;
    }

    // receive each byte of data
    for ( int x = 0; x < dataLength; ++x )
    {
	if ( mlReadByte( &data[ x ] ) == ERROR )
	{
	    return ERROR;
	}
    }

    // done
    return OK;
}


//
// 
//
int
MLBusMaster::mlWriteFunction( int function, int dataLength, const BYTE* data )
{
    // send command
    if ( mlWriteByte( (const BYTE)function ) == ERROR )
    {
	return ERROR;
    }

    // send each byte of data
    for ( int x = 0; x < dataLength; ++x )
    {
	if ( mlWriteByte( data[ x ] ) == ERROR )
	{
	    return ERROR;
	}
    }

    // done
    return OK;
}


//
// MicroLAN(tm) read data phase, NO RESET
//
int
MLBusMaster::mlReadData( int dataLength, BYTE* data )
{
    // receive each byte of data
    for ( int x = 0; x < dataLength; ++x )
    {
	if ( mlReadByte( &data[ x ] ) == ERROR ) return ERROR;
    }

    // done
    return OK;
}


//
// MicroLAN(tm) write data phase, NO RESET
//
int
MLBusMaster::mlWriteData( int dataLength, const BYTE* data )
{
    // send each byte of data
    for ( int x = 0; x < dataLength; ++x )
    {
	if ( mlWriteByte( data[ x ] ) == ERROR ) return ERROR;
    }

    // done
    return OK;
}


//
// complete MicroLAN(tm) function transaction
//
int
MLBusMaster::mlCommand( int commandLength, const BYTE* command,
		        int dataLength, BYTE* data )
{
    int x = 0;

 top:
    while ( x < COMMAND_RETRIES )
    {
	// MUST reset b4 every command/data cycle
	mlReset();

	// send each byte in command
	int y;
	for ( y = 0; y < commandLength; ++y )
	{
	    if ( mlWriteByte( command[ y ] ) == ERROR )
	    {
		++x; goto top;
	    }
	}

	// receive each byte of data
	for ( y = 0; y < dataLength; ++y )
	{
	    if ( mlReadByte( &data[ y ] ) == ERROR )
	    {
		++x; goto top;
	    }
	}

	break;
    }

    // done
    return (x < COMMAND_RETRIES) ? OK : ERROR;
}


//
// send a 'clock' byte to the bus, return data
//
int
MLBusMaster::mlReadByte( BYTE* data )
{
    BYTE byte = 0;

    // read each bit of data
    for ( int x = 0; x < 8; ++x )
    {
	// send 'clock' bit
	DATA( HI_BIT );
	if ( dataDelay() == ERROR ) return ERROR;

	// read returned bit, add to data
	byte >>= 1;
	if ( DATA() & SAMPLE_POSITION ) byte |= EIGHTH_BIT;
    }

    // done
    *data = byte;
    return OK;
}


//
// send a data byte to the bus
//
int
MLBusMaster::mlWriteByte( const BYTE data )
{
    // send each bit of command
    for ( BYTE mask = 0x01; mask; mask <<= 1 )
    {
	if ( data & mask )	DATA( HI_BIT );
	else			DATA( LO_BIT );

	// consume returned data
	if ( dataDelay() == ERROR ) return ERROR;
	DATA();
    }

    // done
    return OK;
}


//
// send a 'clock' bit to the bus, return data
//
int
MLBusMaster::mlReadBit( void )
{
    // send 'clock' bit
    DATA( HI_BIT );

    // read the returned bit
    if ( dataDelay() == ERROR ) return ERROR;

    // 
    return (DATA() & SAMPLE_POSITION) ? 1: 0;
}


//
// send a data bit to the bus
//
int
MLBusMaster::mlWriteBit( const int data )
{
    // 
    if ( data )	DATA( HI_BIT );
    else	DATA( LO_BIT );

    // consume returned data
    if ( dataDelay() == ERROR ) return ERROR;
    DATA();

    // done
    return OK;
}


//
// 8 bit CRC calculation
// this version assumes an 8 byte buffer
// C algorithm by Don McLane
//
int
MLBusMaster::mlCRC8( BYTE* buffer )
{
    BYTE bit, crc = 0;

    for ( int x = 0; x < 8; ++x )
    {
	 for ( BYTE mask = 0x01; mask; mask <<= 1 )
	 {
	      bit = (buffer[ x ] & mask) ? 1 : 0;
	      bit ^= crc & 1;
	      crc >>= 1;
	      if ( bit ) crc ^= 0x8c;
	 }
    }

    return (crc == buffer[ 8 ]) ? TRUE : FALSE;
}


//
// 
//
int
MLBusMaster::dataDelay( void )
{
#define MAX_SPIN	10000
    for ( int x = MAX_SPIN; !(LSR() & DATA_READY); --x )
    {
	if ( !x )
	{
	    cerr << "dataDelay(): timeout" << endl;
	    delay( TENTH_SECOND ); 
	    return ERROR;
	}
    }

    // 
    return OK;
}


//
// 
//
int
MLBusMaster::delay( int millisecs )
{
    int status;

    struct itimerval value =	// seconds, microseconds
    { { 0L, 0L },		// reload value
      { (long)(millisecs / 1000L), (long)(millisecs % 1000L) * 1000L } };

    // create a mask set with SIGALRM set
    /// FIXME: move delaySet to private data
    sigset_t delaySet;
    sigemptyset( &delaySet );
    sigaddset( &delaySet, SIGALRM );
    sigset_t oldSet;

    // prevent loss of signal while cocking itimer
    sigprocmask( SIG_BLOCK, &delaySet, &oldSet );
    setitimer( ITIMER_REAL, &value, NULL );

    // wait for alarm to fire
    sigsuspend( &oldSet );

    // restore old mask, ie without SIGALRM set
    sigprocmask( SIG_UNBLOCK, &delaySet, (sigset_t*)NULL /**&oldSet*/ );
}


//
// 
//
void
MLBusMaster::delayDone( void )
{
/// cerr << "hit itimer!" << endl;
}
