//
// Copyright 1996 Steve Passe
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for any purpose and without fee is hereby granted, provided
// that the above copyright notice appear in all copies and that both that
// copyright notice and this permission notice appear in supporting
// documentation, and that the name of AUTHORS (Steve Passe and Paul Crook)
// are not be used in advertising or publicity pertaining to distribution of
// the software without specific, written prior permission.  AUTHORS make no
// representations about the suitability of this software for any purpose.
// It is provided "as is" without express or implied warranty.
//
// AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL AUTHORS
// BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
// OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
// CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//
// 1996 Revision
// AUTHORS:     Steve Passe <smp@csn.net>
//

static char rcsid[] = "$Id: mldaemon.cc,v 1.8 1996/03/31 03:09:06 smp Exp smp $";


//
// mldaemon.cc
//

#include <iostream.h>
#include <termios.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <signal.h>
#include <sys/file.h>

#include "mldaemon.hh"

#include "mldevice.hh"
#include "mlds1820.hh"
#include "mlds1920.hh"
#include "mlds2401.hh"
#include "mlds2405.hh"
#include "mlds2407.hh"


static int	romTagCmp( const void* p1, const void* p2 );


//
// 
//
MLDaemon::MLDaemon( void )
{
    _port = -1;

    // find and read the button database
    const char*	s = _defaults.getString( "BINMLKEYMAP" );
    if ( s == NULL )
    {
	cerr << "couldn't find MLKEYMAP" << endl;
	terminate();
    }

    /// if ...
    if ( _mlMap.readMap( (char*)s ) == ERROR )
    {
	cerr << "couldn't find read " << s << endl;
	terminate();
    }

    // clear the device pointer array
    memset( _mlDevices, 0, sizeof( MLDevice* ) * _NUM_MLDEVICES );
}


//
// 
//
MLDaemon::~MLDaemon( void )
{
    mlCloseIO(); 
    cerr << "\nmlDaemon stopped" << endl;
}


//
// initializer function
//
int
MLDaemon::initialize( const char* name,	// process name
		        const char* pathName,	// symbolic path name
		        const char* logName,	// symbolic log name
		        const int port )	// device fd
{
    // call the base class initializer
    int status = RcDaemon::initialize( name, pathName, logName, port );
    if ( status != OK )
    {
	return status;
    }

    // announce ourself
    say( "micro lan deemon started,,,\n" );

    return OK;
}


//
// initialize the MicroLAN(tm) hardware/software
//
int
MLDaemon::deviceInitialize( void )
{
    // get a device channel
    if ( openDevice() == ERROR )
    {
	perror( "deviceInitialize(): openDevice()" );
	return ERROR;
    }

    return _port;
}


//
// cleanup
//
void
MLDaemon::terminate( void )
{
    say( "micro lan deemon stopped,,,\n" );

    RcDaemon::terminate();

    exit( 0 );
}


//
// main loop
//
int
MLDaemon::mlMain( void )
{
    // create instance for each physical device
    MLBusMaster master( _port );
    _master = &master;

    // find (and possibly test) all devices on the bus
    if ( mlDiscover() == ERROR ) return ERROR;

    // loop forever...
    while ( TRUE )
    {
	// wait for data
	int status = waitForData();

#if 0	/// FIXME: remove the unnecessary _deviceFD stuff from ???
	if ( status & DEVICE_DATA )
	{
	    ...
	}
#endif

	// look for messages from a client
	if ( status & CLIENT_DATA )
	{
	    if ( doClient() == ERROR )
	    {
		return ERROR;
	    }
	}
    }
}


//
// find (and possibly test) all devices on the bus
//
int
MLDaemon::mlDiscover( void )
{
    // add object for each physical device
    int l = MLDevice::ROM_LENGTH();
    BYTE* ids = new BYTE[ l * _NUM_MLDEVICES ];
    _numberDevices = _master->searchROM( ids, _NUM_MLDEVICES );
    cerr << "found " << _numberDevices << " devices" << endl;

    int	numberDS1820 = 0;
    int	numberDS1920 = 0;
    int	numberDS2401 = 0;
    int	numberDS2405 = 0;
    int	numberDS2407 = 0;
    for ( int x = 0; x < _numberDevices; ++x )
    {
	switch ( ids[ x * l ] )
	{
	case MLDevice::_MLDS1820_FAMILY_CODE:	++numberDS1820;	break;
	///  MLDevice::_MLDS1920_FAMILY_CODE:	++numberDS1920;	break;
	case MLDevice::_MLDS2401_FAMILY_CODE:	++numberDS2401;	break;
	case MLDevice::_MLDS2405_FAMILY_CODE:	++numberDS2405;	break;
	case MLDevice::_MLDS2407_FAMILY_CODE:	++numberDS2407;	break;
	default:
	    cerr << form( "unknown family: 0x%02x", ids[ x * l ] ) << endl;
	    return ERROR;
	}
    }


    // create an array of pointers to objects for each family
    if ( numberDS1820 > 0 )
    {
	_thermometer = new MLDS1820Thermometer*[ numberDS1820 ];
	cerr << "found " << numberDS1820 << " thermometers" << endl;
    }
    else
    {
	_thermometer = NULL;
    }
    numberDS1820 = 0;
    /// numberDS1920 = 0;

    if ( numberDS2401 > 0 )
    {
	_serialNumber = new MLDS2401SerialNumber*[ numberDS2401 ];
	cerr << "found " << numberDS2401 << " serial numbers" << endl;
    }
    else
    {
	_serialNumber = NULL;
    }
    numberDS2401 = 0;

    if ( numberDS2405 > 0 )
    {
	_switch = new MLDS2405AddressableSwitch*[ numberDS2405 ];
	cerr << "found " << numberDS2405 << " single switches" << endl;
    }
    else
    {
	_switch = NULL;
    }
    numberDS2405 = 0;

    if ( numberDS2407 > 0 )
    {
	_dualSwitch = new MLDS2407AddressableSwitch*[ numberDS2407 ];
	cerr << "found " << numberDS2407 << " dual switches" << endl;
    }
    else
    {
	_dualSwitch = NULL;
    }
    numberDS2407 = 0;

    // storage for rom ID to device pointer structures
    _romTags = new romTag_t[ _numberDevices ];

    // store rom tags, create objects, (test objects)
    for ( int x = 0; x < _numberDevices; ++x )
    {
	// 
	memcpy( _romTags[ x ].rom, &ids[ x * l ], 8 );

	switch ( ids[ x * l ] )
	{

	case MLDevice::_MLDS1820_FAMILY_CODE:
	//   MLDevice::_MLDS1920_FAMILY_CODE:

	    // create an instance
	    _thermometer[ numberDS1820 ]
	      = new MLDS1820Thermometer( *_master, &ids[ x * l ] );

	    // check its voltage
	    if ( _thermometer[ numberDS1820 ]->powerSupply() == FALSE )
	    {
		char buf[ 20 ];
		ROMString( _thermometer[ numberDS1820 ], buf );
		cerr << form( "thermometer missing power: %s", buf ) << endl;
		return ERROR;
	    }
	    _romTags[ x ].device = _thermometer[ numberDS1820 ];
	    ++numberDS1820;
	    break;

	case MLDevice::_MLDS2401_FAMILY_CODE:

	    // create an instance
	    _serialNumber[ numberDS2401++ ]
	      = new MLDS2401SerialNumber( *_master, &ids[ x * l ] );
	    break;

	case MLDevice::_MLDS2405_FAMILY_CODE:

	    // create an instance
	    _switch[ numberDS2405++ ]
	      = new MLDS2405AddressableSwitch( *_master, &ids[ x * l ] );
	    break;

	case MLDevice::_MLDS2407_FAMILY_CODE:

	    // create an instance
	    _dualSwitch[ numberDS2407++ ]
	      = new MLDS2407AddressableSwitch( *_master, &ids[ x * l ] );
	    break;

	default:
	    cerr << form( "unknown family: 0x%02x", ids[ x * l ] ) << endl;
	    exit( 1 );
	}
    }

    // sort the rom tags for lookup
    qsort( (void *)_romTags, _numberDevices, sizeof( romTag_t ), romTagCmp );

    /// FIXME:
    // look for missing devices

    /// FIXME:
    // log & alias unexpected devices

    // recover memory
    delete ids;
}


//
// given a stingname for a device, find a pointer to the object that manages it
//
int
MLDaemon::nameToDevice( char* name, MLDevice*& device )
{
    romTag_t	target, *p;

    // get id from name
    _mlMap.getKey( name, (void*)target.rom );

    // search for the data by id
    p = (romTag_t*)bsearch( &target, (void *)_romTags, _numberDevices,
			   sizeof( romTag_t ), romTagCmp );
    if ( p )
    {
	device = p->device;
	return TRUE;
    }
    else
    {
	return FALSE;
    }
}


//
// static function for use with library functions: qsort()/bsearch()
//
static int
romTagCmp( const void* p1, const void* p2 )
{
    return memcmp( (*(romTag_t*)p1).rom, (*(romTag_t*)p2).rom, 8 );
}


//
// format a ROM id into a string
//
void
MLDaemon::ROMString( MLDevice* device, char* asciiBuffer )
{
    BYTE* idp = device->getDeviceID();

    sprintf( asciiBuffer, "%02x%02x%02x%02x%02x%02x%02x%02x",
	     idp[7], idp[6], idp[5], idp[4],
	     idp[3], idp[2], idp[1], idp[0] );
}


//
// 
//
int
MLDaemon::doClient( void )
{

    // client ready,  process packet
    Tokens& tokens = *new Tokens( getClientMsg(), 2 );

    int count;
    if ( (count = tokens.count()) <= 0 )
    {
	if ( _verbose )
	  cerr << "doClient(): no args" << endl;
	returnFailMsg();
    }

    // 
    else if ( (strcmp( tokens[ 0 ], "rr" ) == 0) && (count == 1) )
    {
	char ascii[ 20 ];
	_master->readROM( ascii );
	returnMessage( ascii );
    }

    // 
    else if ( (strcmp( tokens[ 0 ], "f" ) == 0) && (count == 2) )
    {
	// 
	int d;
	if ( (d = validateDevice( tokens[ 1 ] )) == ERROR )
	{
	    returnFailMsg();
	}

	else
	{
	    // request temp
	    double temp;
	    if ( _thermometer[ d ]->fahrenheit( temp ) == ERROR )
	    {
		cerr << "bad conversion" << endl;
		returnFailMsg();
	    }
	    char buffer[ 128 ];
	    sprintf( buffer, "therm #%d: %f", d, temp );
	    returnMessage( buffer );
	    if ( _verbose ) cerr << buffer << endl;
	}
    }

    // 
    else if ( (strcmp( tokens[ 0 ], "F" ) == 0) && (count == 2) )
    {
	// 
	MLDS1820Thermometer* therm;
	if ( nameToDevice( tokens[ 1 ], therm ) == FALSE )
	{
	    returnFailMsg();
	}

	else
	{
	    // request temp
	    double temp;
	    if ( therm->fahrenheit( temp ) == ERROR )
	    {
		cerr << "bad conversion" << endl;
		returnFailMsg();
	    }
	    char buffer[ 128 ];
	    sprintf( buffer, "therm %s: %f", tokens[1], temp );
	    returnMessage( buffer );
	    if ( _verbose ) cerr << buffer << endl;
	}
    }

    //
    else if ( (strcmp( tokens[ 0 ], "p" ) == 0) && (count == 2) )
    {
	// 
	int d;
	if ( (d = validateDevice( tokens[ 1 ] )) == ERROR )
	{
	    returnFailMsg();
	}

	else
	{
	    // request powersupply state
	    int p = _thermometer[ d ]->powerSupply();
	    char buffer[ 128 ];
	    sprintf( buffer, "therm #%d: power supply: %s",
		     d, (p == 1) ? "yes" : "no" );
	    returnMessage( buffer );
	    if ( _verbose ) cerr << buffer << endl;
	}
    }

    // increment the 'verbose' member data
    else if ( strcmp( tokens[ 0 ], "iv" ) == 0 )
    {
	++_verbose;
	returnOkMsg();
    }
    
    // decrement the 'verbose' member data
    else if ( strcmp( tokens[ 0 ], "dv" ) == 0 )
    {
	--_verbose;
	returnOkMsg();
    }
    
    // kill thyself
    else if ( strcmp( tokens[ 0 ], "kill" ) == 0 )
    {
	terminate();
	exit( 2 );			// should never get here!
    }
    
    // no valid command found
    else
    {
	cerr << "Illegal command: '" << tokens[ 0 ];
	if ( count == 2 ) cerr << " " << tokens[ 1 ];
	cerr << "'" << endl;
	returnFailMsg();
    }

    delete &tokens;

    return OK;
}


//
// qualify device @ _mlDevices[ i ] asto family, etc.
//
int
MLDaemon::validateDevice( const char* i )
{
    int index = atoi( i );
    if ( (index < 1) || (index > _NUM_MLDEVICES) )
    {
	return ERROR;
    }

    return index - 1;
}


//
// open the serial port connected to the MicroLAN(tm)
//
int
MLDaemon::openDevice( void )
{
    struct termios	newTerm;

    if ( _port >= 0 )
    {
	return OK;
    }

    // get the default strings
    const char*	mlFile;
    if ( !(mlFile = _defaults.getString( "MLPORT" )) )
    {
	cerr << "couldn't find MLPORT" << endl;
	return ERROR;
    }

    //
    // this device requires DIRECT BIT control of the port,
    // so we dont open it in the normal fashion
    //
    // mlFile will look like: /dev/ttyd[0-n]
    //
    if ( strncmp( mlFile, PORT_STRING, PORTNUM_INDEX ) )
    {
	cerr << "bad portname, expecting '/dev/ttyd[0-3]': " << mlFile << endl;
	return ERROR;
    }

    // 
    _port = mlFile[ PORTNUM_INDEX ] - '0';
    if ( (_port  < 0) || (_port > 3) )
    {
	cerr << "bad port: " << _port << endl;
	return ERROR;
    }

    // enable direct access to IO space
    mlOpenIO();
    
    // all done
    return OK;
}


//
// direct IO enablement
//
int
MLDaemon::mlOpenIO( void )
{
    if ( (_IOPerm = fopen( "/dev/io", "r" )) == NULL )
    {
	perror( "fopen(): " );
	return 0;
    }
    else
    {
	return 1;
    }
}


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

#if 0

//
// process MicroLAN(tm) data
//
int
MLDaemon::processData( int count, unsigned char* data )
{
    if ( count > MAX_DATA ) return ERROR;

    // store in permanent buffer
    memcpy( _data, data, count );

    switch ( data[ 0 ] )
    {

    // all others
    default:
	cerr << "Got unexpected data: ";
	logData( count, data );
	break;
    }

    // result is number of bytes
    return count;
}


//
// log MicroLAN(tm) data
//
int
MLDaemon::logData( int count, const unsigned char* data )
{
    char buffer[ 128 ];

    formatData( count, data, buffer );
    cerr << buffer << endl;

    return OK;
}


//
// format MicroLAN(tm) data
//
int
MLDaemon::formatData( int count, const unsigned char* data, char* buffer )
{
    buffer[ 0 ] = '\0';

    return OK;
}

#endif
