//
// Copyright 1996 Steve Passe
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for any purpose and without fee is hereby granted, provided
// that the above copyright notice appear in all copies and that both that
// copyright notice and this permission notice appear in supporting
// documentation, and that the name of AUTHORS (Steve Passe and Paul Crook)
// are not be used in advertising or publicity pertaining to distribution of
// the software without specific, written prior permission.  AUTHORS make no
// representations about the suitability of this software for any purpose.
// It is provided "as is" without express or implied warranty.
//
// AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL AUTHORS
// BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
// OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
// CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//
// 1996 Revision
// AUTHORS:     Steve Passe <smp@csn.net>
//

// $Id$

//
// MicroLAN is a trademark Dallas Semiconductor
//

//
// mldevice.hh
//


#ifndef _MLDEVICE_H
#define _MLDEVICE_H

#include <string.h>

#include "util/misc/misc.hh"
#include "mlbusmaster.hh"



class MLDevice
{

 public:

    // constructor for devices
    MLDevice( const BYTE family, MLBusMaster& master )
      : _busMaster( master )
    {
	setFamily( family );
    }

    MLDevice( const BYTE* id, MLBusMaster& master )
      : _busMaster( master )
    {
	setDeviceID( id );
    }

    // 
    BYTE*	getDeviceID( void ) { return _deviceID; }
    void	getDeviceID( BYTE* id )
    		  { memcpy( id, _deviceID, _ROM_LENGTH ); }
    void	setDeviceID( const BYTE* id )
		  { memcpy( _deviceID, id, _ROM_LENGTH ); }

    // 
    int		getFamily( void ) { return _deviceID[ _FAMILY_OFFSET ]; }
    void	setFamily( const int family )
		  { _deviceID[ _FAMILY_OFFSET ] = family; }

    // 
    BYTE*	getSerialNumber( void )
		  { return &_deviceID[ _SERIALNUMBER_OFFSET ]; }
    void	getSerialNumber( BYTE* sn )
		  { memcpy( sn,
			    &_deviceID[ _SERIALNUMBER_OFFSET ],
			    _SERIALNUMBER_LENGTH ); }
    void	setSerialNumber( const BYTE* sn )
		  { memcpy( &_deviceID[ _SERIALNUMBER_OFFSET ],
			    sn, _SERIALNUMBER_LENGTH ); }

    // start a command cycle for 'this' device
    int		addressDevice( void )
		  { return _busMaster.matchROM( _deviceID ); }

    // send a CONTROL command to 'this' device
    int		controlDevice( int cmd )
		  { return _busMaster.mlWriteFunction( cmd ); }

    // 
    int		readDevice( int cmd, int len, BYTE* buf )
		  { return _busMaster.mlReadFunction( cmd, len, buf ); }

    // 
    int		delay( int time ) { return _busMaster.delay( time ); }

    // read one bit from the MicroLAN(tm) bus
    int		readBit( void ) { return _busMaster.mlReadBit(); }

    // calculate 8bit CRC
    int		CRC8( BYTE* b ) { return _busMaster.mlCRC8( b ); }

    // 
    int		getCRC( void ) { return _deviceID[ _CRC_OFFSET ]; }
    void	setCRC( const int crc ) { _deviceID[ _CRC_OFFSET ] = crc; }

    // definition of the 'ROM' section of a device
    static int	ROM_LENGTH( void ) { return _ROM_LENGTH; }
    static int	FAMILY_OFFSET( void ) { return _FAMILY_OFFSET; }
    static int	SERIALNUMBER_OFFSET( void ) { return _SERIALNUMBER_OFFSET; }
    static int	SERIALNUMBER_LENGTH( void ) { return _SERIALNUMBER_LENGTH; }
    static int	CRC_OFFSET( void ) { return _CRC_OFFSET; }

    // family codes of devices
    static const int	MLDS1820_FAMILY_CODE( void ) { return _MLDS1820_FAMILY_CODE; }
    static const int	MLDS1920_FAMILY_CODE( void ) { return _MLDS1920_FAMILY_CODE; }
    static const int	MLDS2401_FAMILY_CODE( void ) { return _MLDS2401_FAMILY_CODE; }
    static const int	MLDS2405_FAMILY_CODE( void ) { return _MLDS2405_FAMILY_CODE; }
    static const int	MLDS2407_FAMILY_CODE( void ) { return _MLDS2407_FAMILY_CODE; }
    static const int	MLBUSMASTER_FAMILY_CODE( void )
					{ return _MLBUSMASTER_FAMILY_CODE; }

    // base device commands
    static int	ROM_READ( void ) { return _ROM_READ; }
    static int	ROM_SKIP( void ) { return _ROM_SKIP; }
    static int	ROM_MATCH( void ) { return _ROM_MATCH; }
    static int	ROM_SEARCH( void ) { return _ROM_SEARCH; }

    enum MLDEVICE_FAMILY_CODES
    {
	_MLDS1820_FAMILY_CODE    = 0x10,
	_MLDS1920_FAMILY_CODE    = 0x10,
	_MLDS2401_FAMILY_CODE    = 0x01,
	_MLDS2405_FAMILY_CODE    = 0x05,
	_MLDS2407_FAMILY_CODE    = 0x12,
	_MLBUSMASTER_FAMILY_CODE = 0xff		// 'homemade' code
    };


 protected:

    // busmaster to execute commands
    MLBusMaster&	_busMaster;

    //
    // constants used by MLDevices
    //

    enum MLDEVICE_ROM_SIZES
    {
	_ROM_LENGTH		= 8,
	_FAMILY_OFFSET		= 0,
	_SERIALNUMBER_OFFSET	= 1,
	_SERIALNUMBER_LENGTH	= 6,
	_CRC_OFFSET		= 7,
    };

    enum MLDEVICE_COMMAND
    {
	_ROM_READ      = 0x33,
	_ROM_SKIP      = 0xcc,
	_ROM_MATCH     = 0x55,
	_ROM_SEARCH    = 0xf0
    };


 private:

    BYTE		_deviceID[ _ROM_LENGTH ];

};


#endif // _MLDEVICE_H
