//
// Copyright 1993,1994,1995,1996 Steve Passe and Paul Crook
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for any purpose and without fee is hereby granted, provided
// that the above copyright notice appear in all copies and that both that
// copyright notice and this permission notice appear in supporting
// documentation, and that the name of AUTHORS (Steve Passe and Paul Crook)
// are not be used in advertising or publicity pertaining to distribution of
// the software without specific, written prior permission.  AUTHORS make no
// representations about the suitability of this software for any purpose.
// It is provided "as is" without express or implied warranty.
//
// AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL AUTHORS
// BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
// OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
// CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//
// 1996 Revision
// AUTHORS:     Steve Passe <smp@csn.net>
//              Paul Crook
//

static char rcsid[] = "$Id$";


//
// mlnamer.cc
//

#include <stdio.h>
#include <iostream.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>

#include "util/misc/misc.hh"
#include "util/misc/keymap.hh"
#include "util/misc/namedstr.hh"
#include "rc/daemon/rcserver.hh"

#include "mlnamer.hh"


static int	verbose = 0;

static int	doArgs( int, char** );
static int	nameKeys( const char* );


main( int argc, char *argv[] )
{
    // get args
    if ( doArgs( argc, argv ) == ERROR ) exit( 1 );

    // get the file paths from the RC defaults file
    NamedStr defaults;
    if ( defaults.readStrings( RCD_STRINGS ) == ERROR )
    {
	cerr << "couldn't load necessary file: " << RCD_STRINGS << endl;
	return BAD;
    }
    const char*	mlKeyMapFile;
    if ( !(mlKeyMapFile = defaults.getString( "MLKEYMAP" )) )
    {
	cerr << "couldn't get MLKEYMAP" << endl;
	return BAD;
    }

    // name appropriate keys
    if ( nameKeys( mlKeyMapFile ) == ERROR )
    {
	return BAD;
    }

    // close shop & go home
    return OK;
}


static int
nameKeys( const char* file )
{
    int		x, n, key;
    char	s[ MAX_INPUT + 1 ];
    char	s2[ MAX_INPUT ];
    char*	p;
    char*	p2;
    char	limbName[ MAX_INPUT ];
    char	twigName[ MAX_INPUT ];
    char	stemName[ MAX_INPUT ];
    char	leafName[ MAX_INPUT ];
    char	romString[ MAX_INPUT ];
    FILE*	fp;
    KeyMap	keyMap;

    // open key file
    if ( (fp = fopen( file, "r" )) == NULL )
    {
	cerr << form( "fopen( %s ): %s", file, strerror( errno ) );
	return ERROR;
    }

    // read keymap, building key names
    key = 0;
    while ( fgets( s, MAX_INPUT, fp ) )
    {
	// skip whitespace
	for ( p = s; isspace( *p ); ++p );

	// ignore comments
	if ( (*p == '\0') || (*p == '#') ) continue;

	// look for 'limb' keyword
	if ( strncmp( p, LIMB_TOKEN, strlen(LIMB_TOKEN) ) == EQUAL )
	{
	    twigName[ 0 ] = stemName[ 0 ] = leafName[ 0 ] = '\0';
	    for ( p += strlen(LIMB_TOKEN)+1; isspace( *p ); ++p );
	    for ( x = 0; isprint( *p ); ++x ) limbName[ x ] = *p++;
	    limbName[ x ] = '\0';
	}

	// look for 'twig' keyword
	else if ( strncmp( p, TWIG_TOKEN, strlen(TWIG_TOKEN) ) == EQUAL )
	{
	    stemName[ 0 ] = leafName[ 0 ] = '\0';
	    for ( p += strlen(TWIG_TOKEN)+1; isspace( *p ); ++p );
	    for ( x = 0; isprint( *p ); ++x ) twigName[ x ] = *p++;
	    twigName[ x ] = '\0';
	}

	// look for 'stem' keyword
	else if ( strncmp( p, STEM_TOKEN, strlen(STEM_TOKEN) ) == EQUAL )
	{
	    leafName[ 0 ] = '\0';
	    for ( p += strlen(STEM_TOKEN)+1; isspace( *p ); ++p );
	    for ( x = 0; isprint( *p ); ++x ) stemName[ x ] = *p++;
	    stemName[ x ] = '\0';
	}

	// look for 'leaf' keyword
	else if ( strncmp( p, LEAF_TOKEN, strlen(LEAF_TOKEN) ) == EQUAL )
	{
	    for ( p += strlen(LEAF_TOKEN)+1; isspace( *p ); ++p );
	    for ( x = 0; isprint( *p ); ++x ) leafName[ x ] = *p++;
	    leafName[ x ] = '\0';
	    ++key;
	}

	// look for 'rom' keyword
	else if ( strncmp( p, ROM_TOKEN, strlen(ROM_TOKEN) ) == EQUAL )
	{
	    for ( p += strlen(ROM_TOKEN)+1; isspace( *p ); ++p );
	    for ( x = 0; isprint( *p ); ++x ) romString[ x ] = *p++;
	    romString[ x ] = '\0';
	    if ( x != 16 )
	    {
		cerr << "bad ROM string length: " << x << endl;
		exit( 1 );
	    }

	    // build name string from components
	    s2[ 0 ] = '\0';
	    if ( limbName[ 0 ] ) { strcat( s2, limbName ); strcat( s2, "." ); }
	    if ( twigName[ 0 ] ) { strcat( s2, twigName ); strcat( s2, "." ); }
	    if ( stemName[ 0 ] ) { strcat( s2, stemName ); strcat( s2, "." ); }
	    strcat( s2, leafName );

	    p2 = new char[ strlen( s2 ) + 1 ];
	    strcpy( p2, s2 );

	    // ROM string is kept 'big-endian' for humans
	    // we strip 'bytes' off low end,
	    // building binary 'little-endian' id as we go
	    BYTE romID[ 8 ];
	    int y;
	    for ( x = 0, y = 14; x < 8; ++x )
	    {
		romID[ x ] = strtol( &romString[ y ], (char**)NULL, 16 );
		romString[ y ] = '\0';
		y -= 2;
	    }

	    // store the key
	    keyMap.putKey( p2, romID, 8 );

	    // return storage to system
	    delete p2;
	}
    }

    // close the input file
    fclose( fp );

    // create the output file
    strcat( strcpy( s, file ), ".bin" );
    keyMap.writeMap( s );

    return OK;
}


//
// process command line arguments
//
static int
doArgs( int argc, char *argv[] )
{
    // process optional arguments
    for ( int index = 1; index < argc; ++index )
    {
	if ( argv[ index ][ 0 ] == '-' )
	{
	    switch ( argv[ index ][ 1 ] )
	    {
	    case 'v':
		++verbose;
	        break;

	    default:
		cerr << "usage: " << argv[ 0 ] << "[-v]\n";
		return ERROR;
	    }
	}
	else break;
    }

    return OK;
}
