//
// Copyright 1993,1994,1995 Steve Passe and Paul Crook
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for any purpose and without fee is hereby granted, provided
// that the above copyright notice appear in all copies and that both that
// copyright notice and this permission notice appear in supporting
// documentation, and that the name of AUTHORS (Steve Passe and Paul Crook)
// are not be used in advertising or publicity pertaining to distribution of
// the software without specific, written prior permission.  AUTHORS make no
// representations about the suitability of this software for any purpose.
// It is provided "as is" without express or implied warranty.
//
// AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL AUTHORS
// BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
// OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
// CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//
// 1995 Revision
// AUTHORS:     Steve Passe <smp@csn.net>
//              Paul Crook
//

static char rcsid[] = "$Id: x10namer.c,v 1.1 1995/08/31 01:33:30 smp Exp smp $";


//
// x10namer.cc
//

#include <stdio.h>
#include <stream.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>

#include "util/misc/misc.hh"
#include "util/misc/udgsocket.hh"
#include "util/misc/namedstr.hh"
#include "util/misc/tokens.hh"
#include "util/misc/keymap.hh"
#include "rc/x10d/x10.hh"
#include "rc/daemon/rcserver.hh"


#define MAX_INPUT	255
#define COMMENT_CHAR	'#'

#define AREA_TOKEN	"AREA"
#define UNIT_TOKEN	"UNIT"
#define FUNCTION_TOKEN	"FUNC"


typedef struct valueTag {
    char*	name;
    short	value;
} VALUETAG;

static VALUETAG valueTags[] = {
  { "A:",			AREA_A },
  { "B:",			AREA_B },
  { "C:",			AREA_C },
  { "D:",			AREA_D },
  { "E:",			AREA_E },
  { "F:",			AREA_F },
  { "G:",			AREA_G },
  { "H:",			AREA_H },
  { "I:",			AREA_I },
  { "J:",			AREA_J },
  { "K:",			AREA_K },
  { "L:",			AREA_L },
  { "M:",			AREA_M },
  { "N:",			AREA_N },
  { "O:",			AREA_O },
  { "P:",			AREA_P },
  { "1:",			KEY_1 },
  { "2:",			KEY_2 },
  { "3:",			KEY_3 },
  { "4:",			KEY_4 },
  { "5:",			KEY_5 },
  { "6:",			KEY_6 },
  { "7:",			KEY_7 },
  { "8:",			KEY_8 },
  { "9:",			KEY_9 },
  { "10:",			KEY_10 },
  { "11:",			KEY_11 },
  { "12:",			KEY_12 },
  { "13:",			KEY_13 },
  { "14:",			KEY_14 },
  { "15:",			KEY_15 },
  { "16:",			KEY_16 },
  { "ALL_UNITS_OFF:",		KEY_ALL_UNITS_OFF },
  { "ALL_LIGHTS_ON:",		KEY_ALL_LIGHTS_ON },
  { "ON:",			KEY_ON },
  { "OFF:",			KEY_OFF },
  { "DIM:",			KEY_DIM },
  { "BRIGHT:",			KEY_BRIGHT },
  { "ALL_LIGHTS_OFF:",		KEY_ALL_LIGHTS_OFF },
  { "EXTENDED_CODE:",		KEY_EXTENDED_CODE },
  { "HAIL_REQUEST:",		KEY_HAIL_REQUEST },
  { "HAIL_ACKNOWLEDGE:",	KEY_HAIL_ACKNOWLEDGE },
  { "PRESET_DIM_LOW:",		KEY_PRESET_DIM_LOW },
  { "PRESET_DIM_HIGH:",		KEY_PRESET_DIM_HIGH },
  { "EXTENDED_DATA:",		KEY_EXTENDED_DATA },
  { "STATUS_EQ_ON:",		KEY_STATUS_EQ_ON },
  { "STATUS_EQ_OFF:",		KEY_STATUS_EQ_OFF },
  { "STATUS_REQUEST:",		KEY_STATUS_REQUEST }
};
const TAGCOUNT = (sizeof( valueTags ) / sizeof( VALUETAG ));


static int	doArgs( int, char** );
static int	nameKeys( char* );
static int	tagCmp( const void*, const void* );


main( int argc, char *argv[] )
{
    // get args
    if ( doArgs( argc, argv ) == ERROR ) return BAD;

    // get the file path from the RC defaults file
    NamedStr defaults;
    if ( defaults.readStrings( RCD_STRINGS ) == ERROR )
    {
	cerr << "couldn't load necessary file: " << RCD_STRINGS << "\n";
	return BAD;
    }
    const char*	x10KeyMapFile;
    if ( !(x10KeyMapFile = defaults.getString( "X10KEYMAP" )) )
    {
	cerr << "couldn't get X10KEYMAP\n";
	return BAD;
    }

    // sort the primitive tags
    qsort( (void *)valueTags, TAGCOUNT,	sizeof( VALUETAG ), tagCmp );

    // name appropriate keys
    if ( nameKeys( (char*)x10KeyMapFile ) == ERROR )
    {
	return BAD;
    }

    // close shop & go home
    return OK;
}


static int
nameKeys( char* file )
{
    int		a, f, x;
    char*	p;
    char*	p2;
    char	s[ MAX_INPUT ];
    char	s2[ MAX_INPUT ];
    char	areaName[ MAX_INPUT ];
    char	unitName[ MAX_INPUT ];

    // open file
    FILE* fp;
    if ( (fp = fopen( file, "r" )) == NULL )
    {
	cerr << form( "fopen( %s ): %s", file, strerror( errno ) );
	return BAD;
    }

    // read file, building key names
    KeyMap keyMap;
    VALUETAG target;
    VALUETAG* t;
    while ( fgets( s, MAX_INPUT, fp ) )
    {

	// break string up into individual words
	Tokens& tokens = *new Tokens( s );

	int cnt;
	if ( !(cnt = tokens.count()) ) continue;

	// skip comments
	if ( tokens[ 0 ][ 0 ] == COMMENT_CHAR ) continue;

	// look for 'area' keyword
	  if ( strcmp( tokens[ 0 ], AREA_TOKEN ) == EQUAL )
	  {
	      // check count
	      if ( cnt < 3 )
	      {
		  cerr << "bad AREA\n";
		  continue;
	      }

	      // get area value
	      target.name = tokens[ 1 ];
	      t = (VALUETAG *)bsearch( &target, (void *)valueTags, TAGCOUNT,
				      sizeof( VALUETAG ), tagCmp );
	      if ( t )
	      {
		  a = t->value;
		  strcpy( areaName, tokens[ 2 ] );
	      }
	      else
	      {
		  cerr << "bad AREA\n";
		  continue;
	      }
	  }

	// look for 'unit' keyword
	else if ( (strcmp( tokens[ 0 ], UNIT_TOKEN ) == EQUAL)
		  || (strcmp( tokens[ 0 ], FUNCTION_TOKEN ) == EQUAL) )
	{
	    // check count
	    if ( cnt < 3 )
	    {
		cerr << "bad UNIT/FUNC\n";
		continue;
	    }

	    // get area value
	    target.name = tokens[ 1 ];
	    t = (VALUETAG *)bsearch( &target, (void *)valueTags, TAGCOUNT,
				    sizeof( VALUETAG ), tagCmp );
	    if ( t )
	    {
		f = t->value;
		strcpy( unitName, tokens[ 2 ] );
	    }
	    else
	    {
		cerr << "bad UNIT\n";
		continue;
	    }

	    // create the custom name
	    sprintf( s2, "%s.%s", areaName, unitName );
	    p2 = new char[ strlen( s2 ) + 1 ];
	    strcpy( p2, s2 );
	    keyMap.putKey( p2, a, f );

	    cout << form( "name: '%s', area: %d, func: %d\n", p2, a, f );
	}
    }

    // close the input file
    fclose( fp );

    // create the output file
    strcat( strcpy( s, file ), ".bin" );
    keyMap.writeMap( s );

    return OK;
}


static int
tagCmp( const void* p1,  const void* p2 )
{
    return strcmp( (*(VALUETAG*)p1).name, (*(VALUETAG*)p2).name );
}


//
// process command line arguments
//
static int
doArgs( int argc, char *argv[] )
{
    // process optional arguments
    for ( int index = 1; index < argc; ++index )
    {
	if ( argv[ index ][ 0 ] == '-' )
	{
	    switch ( argv[ index ][ 1 ] )
	    {
#if 0
	    case 'v':
		++verbose;
		break;
#endif
	    default:
		cerr << "usage: " << argv[ 0 ] << "[-v]\n";
		return ERROR;
	    }
	}
	else break;
    }

    return OK;
}
