//
// Copyright (c) 1996
//	Sean Kelly.  All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 3. All advertising materials mentioning features or use of this software
//    must display the following acknowledgement:
//	This product includes software developed by Sean Kelly.
// 4. Neither the name of the author nor the names of any co-contributors
//    may be used to endorse or promote products derived from this software
//    without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY SEAN KELLY AND CONTRIBUTORS ``AS IS'' AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL SEAN KELLY OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.
//
// tcl interface to x10 daemon
//
// This is a tcl module init that adds one new command to the given
// interpreter:
//
//     x10send [-virtual] commands...
//

#include <string.h>
#include <iostream.h>
#include <errno.h>

#include "tcl.h"

#include "util/misc/misc.hh"
#include "util/misc/udgsocket.hh"
#include "util/misc/keymap.hh"
#include "util/misc/namedstr.hh"
#include "util/misc/tokens.hh"
#include "rc/daemon/rcserver.hh"
#include "rc/x10d/x10daemon.hh"
#include "rc/x10tclsh/tclcontrol.hh"

extern "C" int x10send(ClientData, Tcl_Interp*, int, char**);
extern "C" int TclX10_Init(Tcl_Interp* interp);
extern "C" void endTclX10(ClientData);

static UnixDGSocket* sock;
static KeyMap* x10Map;

void endTclX10(ClientData) {
      if (sock) delete sock;
      if (x10Map) delete x10Map;
}

int TclX10_Init(Tcl_Interp* interp) {
      if (Tcl_CreateCommand(interp, "x10send", x10send, 0, endTclX10) == 0)
	    return TCL_ERROR;

      NamedStr defaults;
      if (defaults.readStrings(RCD_STRINGS) == ERROR) {
	    cerr << "couldn't load necessary file: " << RCD_STRINGS << "\n";
	    return TCL_ERROR;
      }
      const char* myPath;
      const char* itsPath;
      if (! (myPath = defaults.getString("RCD_DG_TMP")) || ! (itsPath = defaults.getString("X10D_DG_PATH"))) {
	    cerr << "couldn't get socket paths\n";
	    return TCL_ERROR;
      }
      // create a socket
      sock = new UnixDGSocket((char*)myPath, (char*)itsPath);
      if (sock->getFD() == ERROR) {
	    delete sock;
	    return TCL_ERROR;
      }

      // Get the mapping.  If this fails, oh well.
      x10Map = new KeyMap;
      const char* s;
      if (s = defaults.getString("BINX10KEYMAP")) {
	    if (x10Map->readMap((char*)s) == ERROR)
		  cerr << form( "couldn't read %s\n", s );
      } else
	    cerr << "couldn't find X10KEYMAP\n";

      return TCL_OK;
}

int x10send(ClientData, Tcl_Interp* interp, int argc, char** argv) {
      u_char buf[CNTL_BSIZE];

      // Check number of args
      if (argc <= 1) {
	    interp->result = "wrong # args; usage: [-virtual] <command>...";
	    return TCL_ERROR;
      }
      
      // See if user specified -virtual
      int start;
      if (strcmp(argv[1], "-virtual") == 0) {
	    buf[RC_CMND] = X10VIRTUAL;
	    start = 2;
      } else {
	    buf[RC_CMND] = X10EXEC;
	    start = 1;
      }
      
      // Set the repeat count.
      buf[X10_RCNT] = 2;

      // Map the tokens
      for (int i = start; i < argc; ++i) {
	    if (x10Map->getKey(argv[i], buf[X10_AREA], buf[X10_UNIT]) == FALSE) {
		  sprintf(interp->result, "key not found: %s", argv[i]);
		  return TCL_ERROR;
	    }
	    
	    // Send it.
	    bool resend = true;
	    do {
		  switch (sock->send((char*) buf, CNTL_BSIZE)) {
			case CNTL_BSIZE:
			      resend = false;
			      break;
			case -1:
			      if (errno == ENOBUFS) continue;
			      sprintf(interp->result, "cannot send: %s", strerror(errno));
			      return TCL_ERROR;
			default:
			      interp->result = "bad send to daemon";
			      return TCL_ERROR;
		  }
	    } while (resend);
      }
      
      // Wow.  Made it!
      return TCL_OK;
}
