//
// Copyright 1993,1994,1995 Steve Passe and Paul Crook
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for any purpose and without fee is hereby granted, provided
// that the above copyright notice appear in all copies and that both that
// copyright notice and this permission notice appear in supporting
// documentation, and that the name of AUTHORS (Steve Passe and Paul Crook)
// are not be used in advertising or publicity pertaining to distribution of
// the software without specific, written prior permission.  AUTHORS make no
// representations about the suitability of this software for any purpose.
// It is provided "as is" without express or implied warranty.
//
// AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL AUTHORS
// BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
// OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
// CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//
// 1995 Revision
// AUTHORS:     Steve Passe <smp@csn.net>
//              Paul Crook
//

static char rcsid[] = "$Id: keymap.cc,v 1.2 1996/03/02 07:50:09 smp Exp smp $";

#define OLDIRxxx

//
// keymap.cc
//


#include <stdio.h>
#include <stream.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#include "misc.hh"
#include "keymap.hh"

#define MAX_NAME_SIZE	255


static int keyCmp( const void* p1, const void* p2 );


int
KeyMap::readMap( char* file )
{
    FILE*	fp;
    char*	p;
    char	name[ MAX_NAME_SIZE ];

    // open file
    if ( (fp = fopen( file, "r" )) == NULL ) return ERROR;

    // determine the number of entries
    if ( fread( &keyCount, sizeof( keyCount ), 1, fp ) != 1 )
    {
	return ERROR;
    }

    // allocate storage for all the keys
    keyData = new keydata_t[ keyCount ];

    // read the entries
    for ( int x = 0; x < keyCount; ++x )
    {
	// read & store the name string
	fgets( name, MAX_NAME_SIZE, fp );
	if ( p = strrchr( name, '\n' ) ) *p = '\0';
	KD_NAME( x ) = new char[ strlen( name ) + 1 ];
	strcpy( KD_NAME( x ), name );

	// read the length of the raw data
	int l;
	fread( &l, KD_LSIZE, 1, fp );
	KD_LENGTH( x ) = l;

	// read the raw data
	KD_DATA( x ) = new u_char[ KD_DSIZE * KD_LENGTH( x ) ];
	fread( KD_DATA( x ), KD_DSIZE, KD_LENGTH( x ), fp );
    }

    fclose( fp );

    return OK;
}


int
KeyMap::writeMap( char* file )
{
    FILE* fp;

    // sort the data
    qsort( (void *)keyData, keyCount, sizeof( keydata_t ), keyCmp );

    // open the binary version of the file
    if ( (fp = fopen( file, "w" )) == NULL )
    {
	cerr << form( "writeMap( %s ): %s", file, strerror( errno ) );
	return ERROR;
    }

    // write the count to the file
    fwrite( &keyCount, sizeof( keyCount ), 1, fp );

    // write the records to the file
    for ( int x = 0; x < keyCount; ++x )
    {
	// write the name string
	fputs( KD_NAME( x ), fp );
	fputc( '\n', fp );

	// write the length of the raw data
	int l = KD_LENGTH( x );
	fwrite( &l, KD_LSIZE, 1, fp );

	// write the raw data
	fwrite( KD_DATA( x ), KD_DSIZE, KD_LENGTH( x ), fp );
    }

    fclose( fp );

    return OK;
}


//
// x10 version
//
int
KeyMap::getKey( char* s, X10_DTYPE& a, X10_DTYPE& f )
{
    int		key;
    char	*ptr;
    keydata_t	target, *p;

    // Ann thru Pnn
    if ( (*s >= 'A') && (*s <= 'P') && (isdigit(*(s+1))) )
    {
	a = *s++ - 'A';
	key = strtol( s, &ptr, 10 );
	if ( (s == ptr) || (key < 1) || (key > 32) ) return FALSE;
	f = key - 1;	// zero based
    }

    // real names
    else
    {
	if ( keyCount == 0 ) return FALSE;

	// search for data by name
	target.name = s;
	p = (keydata_t*)bsearch( &target, (void *)keyData, keyCount,
			       sizeof( keydata_t ), keyCmp );
	if ( p )
	{
	    // return the values
	    a = ((X10_DTYPE*)p->data)[ 0 ];
	    f = ((X10_DTYPE*)p->data)[ 1 ];
	}
	else
	{
	    return FALSE;
	}
    }

    return TRUE;
}


//
// x10 version
//
KeyMap::putKey( char* s, u_char a, u_char f )
{
    if ( keyCount == 0 )
    {
	keyData = new keydata_t[ X10_KEY_NUM ];
    }

    // store the name
    KD_NAME( keyCount ) = new char[ strlen( s ) + 1 ];
    strcpy( KD_NAME( keyCount ), s );

    // store the data
    KD_DATA( keyCount ) = new X10_DTYPE[ 3 ];
    KD_LENGTH( keyCount ) = 2;
    X10_DATA( keyCount, 0 ) = a;
    X10_DATA( keyCount, 1 ) = f;

    // bump count
    ++keyCount;

    return OK;
}


//
// ir version
//
int
KeyMap::getKey( char* name, IR_DTYPE* data )
{
    int		key;
    char	*ptr;
    keydata_t	target, *p;

    if ( keyCount == 0 ) return FALSE;

    // search for the data by name
    target.name = name;
    p = (keydata_t*)bsearch( &target, (void *)keyData, keyCount,
			   sizeof( keydata_t ), keyCmp );
    if ( p )
    {
	memcpy( data, p->data, p->length );
    }
    else
    {
	return FALSE;
    }

    return TRUE;
}


//
// ir version
//
KeyMap::putKey( char* name, IR_DTYPE* data )
{
    if ( keyCount == 0 )
    {
	// allocate an array of data items
	keyData = new keydata_t[ IR_KEY_NUM ];
    }

    // store the name
    KD_NAME( keyCount ) = new char[ strlen( name ) + 1 ];
    strcpy( KD_NAME( keyCount ), name );

    // store the raw data length
    KD_LENGTH( keyCount ) = (*data + 1) * IR_DSIZE;

    // store the data
    KD_DATA( keyCount ) = new IR_DTYPE[ KD_LENGTH( keyCount ) ];
    memcpy( KD_DATA( keyCount ), data, KD_LENGTH( keyCount ) );

    // bump the count
    ++keyCount;

    return OK;
}


//
// x10 version
//
int
KeyMap::getName( const char*& s, u_char a, u_char f )
{
    for ( int x = 0; x < keyCount; ++x )
    {
	if ( (X10_DATA( x, 0 ) == a) && (X10_DATA( x, 1 ) == f) )
	{
	    s = KD_NAME( x );
	    return TRUE;
	}
    }

    return FALSE;
}


//
// generic version, COPY keys data to data
//
int
KeyMap::getKey( char* name, void* data )
{
    int		key;
    char	*ptr;
    keydata_t	target, *p;

    if ( keyCount == 0 ) return FALSE;

    // search for the data by name
    target.name = name;
    p = (keydata_t*)bsearch( &target, (void *)keyData, keyCount,
			   sizeof( keydata_t ), keyCmp );
    if ( p )
    {
	memcpy( data, p->data, p->length );
    }
    else
    {
	return FALSE;
    }

    return TRUE;
}


//
// generic version, return POINTER to data and data length
//
int
KeyMap::getKey( char* name, void*& data, int& length )
{
    int		key;
    char	*ptr;
    keydata_t	target, *p;

    if ( keyCount == 0 ) return FALSE;

    // search for the data by name
    target.name = name;
    p = (keydata_t*)bsearch( &target, (void *)keyData, keyCount,
			   sizeof( keydata_t ), keyCmp );
    if ( p )
    {
	data = &p->data;
	length = p->length;
    }
    else
    {
	return FALSE;
    }

    return TRUE;
}


//
// generic version, record name + data[ length ]
//
KeyMap::putKey( char* name, void* data, int length )
{
    if ( keyCount == 0 )
    {
	// allocate an array of data items
	keyData = new keydata_t[ 512 /* FIXME: no magic! */ ];
    }

    // store the name
    KD_NAME( keyCount ) = new char[ strlen( name ) + 1 ];
    strcpy( KD_NAME( keyCount ), name );

    // store the raw data length
    KD_LENGTH( keyCount ) = length;

    // store the data
    KD_DATA( keyCount ) = new BYTE[ KD_LENGTH( keyCount ) ];
    memcpy( KD_DATA( keyCount ), data, KD_LENGTH( keyCount ) );

    // bump the count
    ++keyCount;

    return OK;
}


//
// generic version
//
int
KeyMap::getName( const char*& name, BYTE* data )
{
    for ( int x = 0; x < keyCount; ++x )
    {
	if ( memcmp( KD_DATA( x ), data, KD_LENGTH( x ) ) == 0 )
	{
	    name = KD_NAME( x );
	    return TRUE;
	}
    }

    name = NULL;
    return FALSE;
}


static int
keyCmp( const void* p1, const void* p2 )
{
    return strcmp( (*(keydata_t*)p1).name, (*(keydata_t*)p2).name );
}
