//
// Copyright 1996 Steve Passe
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for any purpose and without fee is hereby granted, provided
// that the above copyright notice appear in all copies and that both that
// copyright notice and this permission notice appear in supporting
// documentation, and that the name of AUTHORS (Steve Passe and Paul Crook)
// are not be used in advertising or publicity pertaining to distribution of
// the software without specific, written prior permission.  AUTHORS make no
// representations about the suitability of this software for any purpose.
// It is provided "as is" without express or implied warranty.
//
// AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL AUTHORS
// BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
// OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
// CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//
// 1996 Revision
// AUTHORS:     Steve Passe <smp@csn.net>
//

static char rcsid[] = "$Id$";


//
// rccontrol.cc
//

#include <stdio.h>
#include <stream.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>

#include "util/misc/misc.hh"
#include "util/misc/namedstr.hh"
#include "util/misc/udgsocket.hh"
#include "rc/daemon/connection.hh"
#include "rc/daemon/rcserver.hh"

#include "rccontrol.hh"

static int		doArgs( int argc, char** argv, RcControl& rcc );


//
// a proof-of-concept remote control client
// supports async status from HCSII and X-10 daemons
//
// Note: needs work in the ERROR detection department!
//

main( int argc, char* argv[] )
{
    // create the (ONE and ONLY ONE) instance of the daemon object
    RcControl rcc( argv[ 0 ] );

    // get args
    if ( doArgs( argc, argv, rcc ) == ERROR )
    {
	exit( 1 );
    }

    // initialize client, NO default daemon, NO logfile
    if ( rcc.initialize() == ERROR )
    {
	cerr << "rccontrol failure!" << endl;
	exit( 1 );
    }

    //
    // do the work here
    //
    return rcc.rcMain();
}


//
// 
//
int
RcControl::rcMain( void )
{
    // make an HCSII command connection
    Connection*	h2Cmnd;
    int h2CmndFD;
    makeUDGConnection( "HCS2D_DG_PATH", h2Cmnd, h2CmndFD );

    // make an HCSII event connection, get local address for later use
    Connection*	h2Event;
    int h2EventFD;
    char h2EventAddr[ UDG_TMPADDR_SIZE ];
    makeUDGConnection( "HCS2D_DG_PATH", h2Event, h2EventFD, h2EventAddr );

    // make an X10 command connection
    Connection*	x10Cmnd;
    int x10CmndFD;
    makeUDGConnection( "X10D_DG_PATH", x10Cmnd, x10CmndFD );

    // make an X10 event connection, get local address for later use
    Connection*	x10Event;
    int x10EventFD;
    char x10EventAddr[ UDG_TMPADDR_SIZE ];
    makeUDGConnection( "X10D_DG_PATH", x10Event, x10EventFD, x10EventAddr );

    // main loop
    while ( TRUE )
    {
	// wait for data
	int dataAvailable = waitOnData();

	// data from client/user interface
	if ( clientDataReady() )
	{
	    if ( doClient( h2Cmnd, h2EventAddr, x10Cmnd, x10EventAddr)
		  == FALSE ) break;    // reached EOF on client input
	}

	// message from the HCSII daemon
	if ( dataReady( h2CmndFD ) ) doHcs2Cmnd( h2Cmnd );

	// message from HCSII event channel
	if ( dataReady( h2EventFD ) ) doHcs2Event( h2Event );

	// message from the X10 daemon
	if ( dataReady( x10CmndFD ) ) doX10Cmnd( x10Cmnd );

	// message from X10 event channel
	if ( dataReady( x10EventFD ) ) doX10Event( x10Event );

    }  // endOfWhile( TRUE )

    // done
    return OK;
}


//
// 
//
int
RcControl::doClient( Connection* h2Cmnd, const char* h2EventAddr,
		     Connection* x10Cmnd, const char* x10EventAddr )
{
    // get input from stdin
    char str[ MAX_INPUT ];
    if ( fgets( str, MAX_INPUT, stdin ) == NULL ) return FALSE;

    // remove WS, look for comments, toss 'empty' lines
    char* p = str;
    while ( isspace( *p ) ) ++p;
    if ( (*p == COMMENT_CHAR) || (*p == '\0') )
    {
	return TRUE;
    }

    // HCSII notify commands
    // format: <hm> <on|off> <i|o.#>
    // eg. "hm on i.14" turns on status reports for change of input #14
    // once an hcs2d namer is built <i|o.#> will become <descriptive.name>
    if ( strncmp( p, "hm on", 5 ) == 0 )
    {
	char ss[ 1024 ];
	char* ptr = strrchr( p, '\n' );
	if ( ptr ) *ptr = '\0';
	sprintf( ss, "monitor %s %s", &p[ 3 ], h2EventAddr );
	h2Cmnd->send( ss );
	h2Cmnd->recv( str );
	cout << ss << ": " << str << endl;
    }
    else if ( strncmp( p, "hm off", 6 ) == 0 )
    {
	char ss[ 1024 ];
	char* ptr = strrchr( p, '\n' );
	if ( ptr ) *ptr = '\0';
	sprintf( ss, "monitor %s %s", &p[ 3 ], h2EventAddr );
	h2Cmnd->send( ss );
	h2Cmnd->recv( str );
	cout << ss << ": " << str << endl;
    }

    // X10 notify commands
    // format: <xm> <on|off> <area.unit>
    // eg. "xm on office.light" turns on status reports for
    //      changes in office light, ie on or off
    else if ( strncmp( p, "xm on", 5 ) == 0 )
    {
	char ss[ 1024 ];
	char* ptr = strrchr( p, '\n' );
	if ( ptr ) *ptr = '\0';
	sprintf( ss, "monitor %s %s", &p[ 3 ], x10EventAddr );
	x10Cmnd->send( ss );
	x10Cmnd->recv( str );
	cout << ss << ": " << str << endl;
    }
    else if ( strncmp( p, "xm off", 6 ) == 0 )
    {
	char ss[ 1024 ];
	char* ptr = strrchr( p, '\n' );
	if ( ptr ) *ptr = '\0';
	sprintf( ss, "monitor %s %s", &p[ 3 ], x10EventAddr );
	x10Cmnd->send( ss );
	x10Cmnd->recv( str );
	cout << ss << ": " << str << endl;
    }

    // x10 commands
    else if ( strncmp( p, "xon", 3 ) == 0 )
    {
	x10Cmnd->send( "emit office.light office.on" );
    }
    else if ( strncmp( p, "xoff", 4 ) == 0 )
    {
	x10Cmnd->send( "emit office.light office.off" );
    }

    else
    {
	cerr << "huh???" << endl;
    }

    // done
    return TRUE;
}


//
// 
//
int
RcControl::doHcs2Cmnd( Connection* h2Cmnd )
{
    char message[ MAX_DG_MSG ];
    if ( h2Cmnd->recv( message ) < 0 )
    {
	perror( "readDaemonMsg()" );
	return ERROR;
    }

    // process 'status' message
    if ( message[ 0 ] == STATUS_MSG_START )
    {
	cout << form( "h2Cmnd status: %s\n", message );
    }

    else
    {
	cerr << form( "unexpected daemon data: '%s'", message) << endl;
    }

    // done
    return OK;
}


//
// 
//
int
RcControl::doHcs2Event( Connection* h2Event )
{
    char message[ MAX_DG_MSG ];
    if ( h2Event->recv( message ) < 0 )
    {
	perror( "h2Event->recv()" );
	return ERROR;
    }

    // process 'status' message
    if ( message[ 0 ] == STATUS_MSG_START )
    {
	cout << form( "h2Event status: %s\n", message );
    }

    // async data
    else
    {
	Tokens& tokens = *new Tokens( message );

	char s[ 256 ];
	char* p = strchr( tokens[ 0 ], '.' );
	sprintf( s, "The %s is now %s.\n",
		 sensorData[ atoi( ++p ) ].name, tokens[ 1 ] );
	cout << s;

	if ( getVerbose() ) say( s );

	delete &tokens;
    }

    // done
    return OK;
}


//
// 
//
int
RcControl::doX10Cmnd( Connection* x10Cmnd )
{
    char message[ MAX_DG_MSG ];
    if ( x10Cmnd->recv( message ) < 0 )
    {
	perror( "readDaemonMsg()" );
	return ERROR;
    }

    // process 'status' message
    if ( message[ 0 ] == STATUS_MSG_START )
    {
	cout << form( "x10Cmnd status: %s\n", message );
    }

    else
    {
	cerr << form( "unexpected X10 daemon data: '%s'", message) << endl;
    }

    // done
    return OK;
}


//
// 
//
int
RcControl::doX10Event( Connection* x10Event )
{
    char message[ MAX_DG_MSG ];
    if ( x10Event->recv( message ) < 0 )
    {
	perror( "x10Event->recv()" );
	return ERROR;
    }

    // process 'status' message
    if ( message[ 0 ] == STATUS_MSG_START )
    {
	cout << form( "x10Event status: %s\n", message );
    }

    // async data
    else
    {
	Tokens& tokens = *new Tokens( message );

	char s[ 256 ];
	char* p = strchr( tokens[ 0 ], '.' );
	*p++ = '\0';
	sprintf( s, "The %s %s is now %s.\n",
		 tokens[ 0 ], p, tokens[ 1 ] );
	cout << s;

	if ( getVerbose() ) say( s );

	delete &tokens;
    }

    // done
    return OK;
}


//
// process command line arguments
//
static int
doArgs( int argc, char** argv, RcControl& rcc )
{
    // process optional arguments
    for ( int index = 1; index < argc; ++index )
    {
	if ( argv[ index ][ 0 ] == '-' )
	{
	    switch ( argv[ index ][ 1 ] )
	    {
	    case 'v':
		rcc.incVerbose();
	        break;

	    default:
		cerr << "usage: %s [-v]" << argv[ 0 ] << endl;
		return ERROR;
	    }
	}
	else break;
    }

    return OK;
}
